package org.expeditee.io;

import java.awt.Image;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.List;

import org.expeditee.gui.Frame;
import org.expeditee.gui.UserSettings;
import org.expeditee.items.Item;
import org.expeditee.items.Picture;
import org.expeditee.items.Text;

import com.lowagie.text.Document;
import com.lowagie.text.DocumentException;
import com.lowagie.text.FontFactory;
import com.lowagie.text.Paragraph;
import com.lowagie.text.pdf.PdfWriter;

/**
 * Writes out a tree in a linear format.
 * @author root
 *
 */
public class PDFWriter extends DefaultTreeWriter {

	private Document _pdfDocument;

	public PDFWriter() {
		_pdfDocument = new Document();
	}

	@Override
	protected void initialise(Frame start) throws IOException {
		_format = ".pdf";
		super.initialise(start);
		try {
			PdfWriter
					.getInstance(_pdfDocument, new FileOutputStream(_filename));
			_pdfDocument.open();
			_pdfDocument.addCreationDate();
			_pdfDocument.addAuthor(UserSettings.Username);
			_pdfDocument.addCreator("Expeditee");
		} catch (DocumentException e) {
			e.printStackTrace();
			throw new IOException(e.getMessage());
		}
	}

	@Override
	protected void writeTitle(Text toWrite, List<Item> items)
			throws IOException {
		_pdfDocument.addTitle(toWrite.getText());
		writeText(toWrite);
	}

	@Override
	protected String finaliseTree() throws IOException {
		_pdfDocument.close();
		return super.finaliseTree();
	}

	@Override
	protected void writeStartLink(Item linker) throws IOException {
	}

	@Override
	protected void writeEndLink(Item linker) throws IOException {
	}

	@Override
	protected void writeText(Text text) throws IOException {
		try {

			_pdfDocument.add(new Paragraph(text.getText(), FontFactory.getFont(
					Conversion.getPdfFont(text.getFamily()), text.getSize(),
					text.getPaintFont().getStyle(), text.getColor())));
		} catch (DocumentException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	@Override
	protected void writePicture(Picture pic) throws IOException {
		Image image = pic.getCroppedImage();
		try {
			_pdfDocument.add(com.lowagie.text.Image.getInstance(image, null));
		} catch (DocumentException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
}
