package org.expeditee.network;

import java.util.HashMap;
import java.util.Map;

import org.expeditee.gui.AttributeValuePair;
import org.expeditee.gui.Frame;
import org.expeditee.gui.MessageBay;
import org.expeditee.items.Text;
import org.net.p2p.Connection;
import org.net.p2p.Protocol;
import org.net.p2p.jnmp2p;


public class PeerToPeer {
	public static boolean _autoConnect = false;

	private static PeerToPeer _theSession = null;

	private Map<String, AttributeValuePair> _peers;
	
	private Map<String, Connection> _peerConnections;

	private Boolean _bConnecting;

	public static void init(Frame settingsFrame) {
		if (settingsFrame == null)
			return;

		if (_theSession == null)
			_theSession = new PeerToPeer(settingsFrame);
	}
	
	private PeerToPeer(Frame settingsFrame) {
		_bConnecting = false;

		_peers = new HashMap<String, AttributeValuePair>();

		// Set the settings
		for (Text item : settingsFrame.getBodyTextItems(false)) {
			if (item.getText().toLowerCase().trim().equals("autoconnect")) {
				_autoConnect = true;
				continue;
			}
			AttributeValuePair avp = new AttributeValuePair(item.getText());
			if (!avp.hasPair())
				continue;
			String attributeFullCase = avp.getAttribute();
			String attribute = attributeFullCase.toLowerCase();
			
			_peers.put(attribute, avp);
		}

		
		try {
			if (_autoConnect) {
				connectThreaded();
			}
		} catch (Exception e) {
			MessageBay.errorMessage("Error in PeerToPeer setup");
		}
	}

	/**
	 * Attempts to connect with peers.
	 */
	public static void connect() {
		if (_theSession._bConnecting) {
			MessageBay.errorMessage("Already connecting to peers");
			return;
		} else if (_theSession != null) {
			_theSession.connectThreaded();
		}
	}

	private void connectThreaded() {
		Thread t = new ConnectThread(this);
		t.start();
	}

	public synchronized void connectToPeers() {
		_bConnecting = true;
		
		//Connect to peers
		Protocol p = new Protocol(new Object());
		
		jnmp2p jnm = new jnmp2p(p, 3000);
		
		for(AttributeValuePair avp: _peers.values()){
			_peerConnections.put(avp.getAttribute().toLowerCase(), jnm.connect(avp.getValue()));
		}
		
		_bConnecting = false;
	}
}
